<?php namespace Myriad\NotifyClient;

use function array_key_exists;
use Myriad\NotifyClient\Client\Config;
use Myriad\NotifyClient\Client\FluentApi\NotificationGroupFluentApi;
use Myriad\NotifyClient\Client\FluentApi\NotificationFluentApi;
use Myriad\NotifyClient\Client\HttpWrapper;
use function str_random;

function softGet($arr, $key, $default = null) {
  return array_key_exists($key, $arr) ? $arr[$key] : $default;
}

/**
 * Generates a random notification group id
 * @return string
 */
function generateUID() {
  return str_random(96);
}

/**
 * Class Client
 *
 * This class is the main entry point for the Notification Service Client SDK.
 *
 * @package Myriad\NotifyClient
 */
class Client {


  /**
   * Delegated to make HTTP requests, builds the uri, processes the responses
   */
  private $http;

  /**
   * Holds all the configuration data for the Client. It is passed to internally-used
   * classes as needed.
   *
   * @var Config
   */
  private $config;

  /**
   * @var Client\Helpers\FilterHelper
   */
  private $filterHelper;

  /**
   * @var Client\Helpers\NotificationGroupHelper
   */
  private $notificationGroupHelper;

  /**
   * @var Client\Helpers\NotificationHelper
   */
  private $notificationHelper;

  /**
   * @var Client\Helpers\NotificationGroupRegistrantHelper
   */
  private $registrantHelper;

  function __construct(Config $config, HttpInterface $http) {
    $this->config = $config;
    $this->filterHelper = new Client\Helpers\FilterHelper();
    $this->notificationGroupHelper = new Client\Helpers\NotificationGroupHelper($config);
    $this->notificationHelper = new Client\Helpers\NotificationHelper($config);
    $this->registrantHelper = new Client\Helpers\NotificationGroupRegistrantHelper($config);

    $this->http = new HttpWrapper($config, $http);

  }

  /**
   * Creates a randomly-generated notification group id. You should generate and save an id with each
   * new user created.
   * @return string
   */
  public static function generateUID() {
    return generateUID();
  }

  /**
   * Fluent API for a notificationGroup-related request. A 'uid' can be passed for use with subsequent methods.
   * This can be useful when building a query object to make multiple requests for the same notificationGroup.
   * @param null $uid
   * @return NotificationGroupFluentApi
   */
  public function notificationGroups($uid = null) {
    return new NotificationGroupFluentApi($this, $uid);
  }

  /**** Non-fluent API for managing notification groups ****/

  public function createUpdateNotificationGroup($notificationGroup, $uid = null) {
    $h = $this->notificationGroupHelper;
    $data = $h->makeArray($notificationGroup);
    $h->validate($data);
    $uid = $uid ? $uid : softGet($data, 'uid', $this->generateUID());
    return $h->hydrateItem($this->http->put($h->uri($uid), $data));
  }

  public function queryNotificationGroups($filter, $sort = null, $limit = null) {
    $h = $this->notificationGroupHelper;
    $filter = $this->filterHelper->makeArray($filter);
    return $h->hydrateCollection(
      $this->http->post(
        $h->uri("query"),
        array_filter(
          [
            'filter' => $filter,
            'sort' => $sort,
            'limit' => $limit
          ]
        )
      )
    );
  }

  public function getNotificationGroup($uid) {
    $h = $this->notificationGroupHelper;
    return $h->hydrateItem($this->http->get($h->uri($uid)));
  }

  public function deleteNotificationGroup($uid) {
    $h = $this->notificationGroupHelper;
    return $this->http->delete($h->uri($uid));
  }

  public function addChannelsToNotificationGroup($uid, array $channels) {
    $h = $this->notificationGroupHelper;
    $data = $h->makeArray(['channels' => $channels]);
    $h->validate($data);
    return $h->hydrateItem($this->http->put($h->uri("$uid/channels"), $data));
  }

  public function removeChannelsFromNotificationGroup($uid, array $channels) {
    $h = $this->notificationGroupHelper;
    $data = $h->makeArray(['channels' => $channels]);
    $h->validate($data);
    return $h->hydrateItem($this->http->delete($h->uri("$uid/channels"), $data));
  }

  public function addUpdateCustomToNotificationGroup($uid, array $custom) {
    $h = $this->notificationGroupHelper;
    $data = $h->makeArray(['custom' => $custom]);
    $h->validate($data);
    return $h->hydrateItem($this->http->put($h->uri("$uid/custom"), $data));
  }

  public function removeCustomFromNotificationGroup($uid, array $custom) {
    $h = $this->notificationGroupHelper;
    $data = $h->makeArray(['custom' => $custom]);
    $h->validate($data);
    return $h->hydrateItem($this->http->delete($h->uri("$uid/custom"), $data));
  }

  /**** Non-fluent API for managing notification group registrants *****/

  public function allRegistrants($uid) {
    $h = $this->registrantHelper;
    return $h->hydrateCollection($this->http->get($h->uri($uid)));
  }

  public function getRegistrant($uid, $registrantSlug) {
    $h = $this->registrantHelper;
    return $h->hydrateItem($this->http->get($h->uri($uid, $registrantSlug)));
  }

  public function createUpdateRegistrant($uid, $registrant, $registrantSlug = null) {
    $h = $this->registrantHelper;
    $registrant = $h->makeArray($registrant);
    $h->validate($registrant);
    $registrantSlug = $registrantSlug ? $registrantSlug : softGet($registrant, 'slug');
    return $h->hydrateItem($this->http->get($h->uri($uid, $registrantSlug), $registrant));
  }

  public function deleteRegistrant($uid, $registrantSlug) {
    $h = $this->registrantHelper;
    return $this->http->delete($h->uri($uid, $registrantSlug));
  }

  public function addSubscriptionsToRegistrant($uid, $registrantSlug, array $subscriptions) {
    $h = $this->registrantHelper;
    $registrant = ['subscriptions' => $subscriptions];
    $h->validate($registrant);
    return $h->hydrateItem(
      $this->http->put($h->uri($uid, "$registrantSlug/subscriptions"), $registrant)
    );
  }

  public function removeSubscriptionsFromRegistrant($uid, $registrantSlug, array $subscriptions) {
    $h = $this->registrantHelper;
    $registrant = ['subscriptions' => $subscriptions];
    $h->validate($registrant);
    return $h->hydrateItem(
      $this->http->delete($h->uri($uid, "$registrantSlug/subscriptions"), $registrant)
    );
  }

  public function addUpdateCustomToRegistrant($uid, $registrantSlug, $custom) {
    $h = $this->registrantHelper;
    return $h->hydrateItem(
      $this->http->put($h->uri($uid, "$registrantSlug/custom"), ['custom' => $custom])
    );
  }

  public function removeCustomFromRegistrant($uid, $registrantSlug, $custom) {
    $h = $this->registrantHelper;
    return $h->hydrateItem(
      $this->http->delete($h->uri($uid, "$registrantSlug/custom"), ['custom' => $custom])
    );
  }


  /**
   * Fluent API for notification-related request. The notificationId parameter is required to use
   * subsequent methods which operate on an individual notification, with the exception of `create`.
   * @param null $notificationId
   * @return NotificationFluentApi
   */
  public function notifications($notificationId = null) {
    return new NotificationFluentApi($this, $notificationId);
  }

  /**** Non-fluent API for managing notifications ****/


  public function queryNotifications($filter, $sort = null, $limit = null) {
    $h = $this->notificationHelper;
    $filter = $this->filterHelper->makeArray($filter);
    return $h->hydrateCollection(
      $this->http->post(
        $h->uri('query'),
        array_filter(
          [
            'filter' => $filter,
            'sort' => $sort,
            'limit' => $limit
          ]
        )
      )
    );
  }

  public function createNotification($notification) {
    $h = $this->notificationHelper;
    $h->maybeValidateBuilder($notification);
    $notification = $h->makeArray($notification);
    return $h->hydrateItem($this->http->post($h->uri(), $notification));
  }

  public function getNotification($notificationId) {
    $h = $this->notificationHelper;
    return $h->hydrateItem($this->http->get($h->uri($notificationId)));
  }

  public function updateNotification($notification, $id = null) {
    $h = $this->notificationHelper;
    $h->maybeValidateBuilder($notification);
    $notification = $h->makeArray($notification);
    return $h->hydrateItem($this->http->put($h->uri($id ? $id : softGet($notification, '_id')), $notification));
  }

  public function deleteNotification($notificationId) {
    $h = $this->notificationHelper;
    return $this->http->delete($h->uri($notificationId));
  }
}
