<?php namespace Myriad\NotifyClient\Utilities;

use ArrayAccess;
use Exception;
use JsonSerializable;

abstract class BasicModel implements ArrayAccess, JsonSerializable {

  protected $attributes = [];

  protected $attrMap = [];

  /**
   * Create a new model instance.
   *
   * @param  array $attributes
   */
  public function __construct(array $attributes = []) {
    $this->hydrate($attributes);
  }

  /**
   * Fill the model with an array of attributes.
   *
   * @param  array $attributes
   * @return $this
   */
  public function hydrate(array $attributes) {
    foreach ($attributes as $key => $value) {
      $this->setAttribute($key, $value);
    }

    return $this;
  }

  /**
   * Convert the model instance to an array.
   *
   * @return array
   */
  public function toArray() {
    return $this->attributes;
  }

  /**
   * Convert the model instance to JSON.
   *
   * @param  int $options
   * @return string
   * @throws \Exception
   */
  public function toJson($options = 0) {
    $json = json_encode($this->jsonSerialize(), $options);

    if (JSON_ERROR_NONE !== json_last_error()) {
      throw new Exception(json_last_error_msg());
    }

    return $json;
  }

  /**
   * Convert the object into something JSON serializable.
   *
   * @return array
   */
  public function jsonSerialize() {
    return $this->toArray();
  }

  /**
   * Dynamically retrieve attributes on the model.
   *
   * @param  string $key
   * @return mixed
   */
  public function __get($key) {
    return $this->getAttribute($key);
  }

  /**
   * Dynamically set attributes on the model.
   *
   * @param  string $key
   * @param  mixed $value
   * @return void
   */
  public function __set($key, $value) {
    $this->setAttribute($key, $value);
  }


  /**
   * Determine if an attribute or relation exists on the model.
   *
   * @param  string $key
   * @return bool
   */
  public function __isset($key) {
    return !is_null($this->getAttribute($key));
  }

  /**
   * Unset an attribute on the model.
   *
   * @param  string $key
   * @return void
   */
  public function __unset($key) {
    unset($this->attributes[$key]);
  }

  /**
   * Handle dynamic static method calls into the method.
   *
   * @param  string $method
   * @param  array $parameters
   * @return mixed
   */
  public static function __callStatic($method, $parameters) {
    return (new static)->$method(...$parameters);
  }

  /**
   * Convert the model to its string representation.
   *
   * @return string
   */
  public function __toString() {
    return $this->toJson();
  }

  /**
   * Create a new instance of the given model.
   *
   * @param  array $attributes
   * @return static
   */
  public function newInstance($attributes = []) {
    return new static((array)$attributes);
  }

  /**
   * Determine if the given attribute exists.
   *
   * @param  mixed $offset
   * @return bool
   */
  public function offsetExists($offset) {
    return isset($this->$offset);
  }

  /**
   * Get the value for a given offset.
   *
   * @param  mixed $offset
   * @return mixed
   */
  public function offsetGet($offset) {
    return $this->$offset;
  }

  /**
   * Set the value for a given offset.
   *
   * @param  mixed $offset
   * @param  mixed $value
   * @return void
   */
  public function offsetSet($offset, $value) {
    $this->$offset = $value;
  }

  /**
   * Unset the value for a given offset.
   *
   * @param  mixed $offset
   * @return void
   */
  public function offsetUnset($offset) {
    unset($this->$offset);
  }

  private function getAttribute($key, $default = null) {
    if (array_key_exists($key, $this->attrMap)) {
      $key = $this->attrMap[$key];
    }

    return $this->arrayGet($this->attributes, $key, $default);
  }

  private function setAttribute($key, $value) {
    if (array_key_exists($key, $this->attrMap)) {
      $key = $this->attrMap[$key];
    }

    $this->arraySet($this->attributes, $key, $value);
  }

  private function arrayGet($array, $key, $default = null)
  {
    if (! is_array($array)) {
      return $default;
    }

    if (is_null($key)) {
      return $array;
    }

    if (array_key_exists($key, $array)) {
      return $array[$key];
    }

    foreach (explode('.', $key) as $segment) {
      if (is_array($array) && array_key_exists($segment, $array)) {
        $array = $array[$segment];
      } else {
        return $default;
      }
    }

    return $array;
  }


  private function arraySet(&$array, $key, $value)
  {
    if (is_null($key)) {
      return $array = $value;
    }

    $keys = explode('.', $key);

    while (count($keys) > 1) {
      $key = array_shift($keys);

      // If the key doesn't exist at this depth, we will just create an empty array
      // to hold the next value, allowing us to create the arrays to hold final
      // values at the correct depth. Then we'll keep digging into the array.
      if (! isset($array[$key]) || ! is_array($array[$key])) {
        $array[$key] = [];
      }

      $array = &$array[$key];
    }

    $array[array_shift($keys)] = $value;

    return $array;
  }
}